# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest
import sympy

from sympy import Rational as Rat

from sppl.sets import EmptySet
from sppl.sets import FiniteNominal
from sppl.sets import FiniteReal
from sppl.sets import Interval
from sppl.sets import inf as oo

from sppl.transforms import Abs
from sppl.transforms import Exp
from sppl.transforms import Identity
from sppl.transforms import Log
from sppl.transforms import Piecewise
from sppl.transforms import Poly
from sppl.transforms import Pow
from sppl.transforms import Radical
from sppl.transforms import Reciprocal
from sppl.transforms import Sqrt

from sppl.transforms import EventAnd
from sppl.transforms import EventFiniteNominal
from sppl.transforms import EventFiniteReal
from sppl.transforms import EventInterval
from sppl.transforms import EventOr

X = Identity("X")
Y = X

# Avoid reporting tests which raise errors:
# pylint: disable=pointless-statement
# pylint: disable=expression-not-assigned

def test_parse_1_open():
    # log(x) > 2
    expr = Log(X) > 2
    event = EventInterval(Log(Y), Interval(2, oo, left_open=True))
    assert expr == event

def test_parse_1_closed():
    # log(x) >= 2
    expr = Log(X) >= 2
    event = EventInterval(Log(Y), Interval(2, oo))
    assert expr == event

def test_parse_2_open():
    # log(x) < 2 & (x < exp(2))
    expr = (Log(X) > 2) & (X < sympy.exp(2))
    event = EventAnd([
        EventInterval(Log(Y), Interval.open(2, oo)),
        EventInterval(Y, Interval.open(-oo, sympy.exp(2)))
    ])
    assert expr == event

def test_parse_2_closed():
    # (log(x) <= 2) & (x >= exp(2))
    expr = (Log(X) >= 2) & (X <= sympy.exp(2))
    event = EventAnd([
        EventInterval(Log(Y), Interval(2, oo)),
        EventInterval(Y, Interval(-oo, sympy.exp(2)))
    ])
    assert expr == event

def test_parse_4():
    # (x >= 0) & (x <= 0)
    expr = (X >= 0) | (X <= 0)
    event = EventInterval(X, Interval(-oo, oo))
    assert expr == event

def test_parse_5_open():
    # (2*x+10 < 4) & (x + 10 > 3)
    expr = ((2*X + 10) < 4) & (X + 10 > 3)
    event = EventAnd([
        EventInterval(Poly(Y, [10, 2]), Interval.open(-oo, 4)),
        EventInterval(Poly(Y, [10, 1]), Interval.open(3, oo)),
    ])
    assert expr == event

def test_parse_5_ropen():
    # (2*x+10 < 4) & (x + 10 >= 3)
    expr = ((2*X + 10) < 4) & (X + 10 >= 3)
    event = EventAnd([
        EventInterval(Poly(Y, [10, 2]), Interval.open(-oo, 4)),
        EventInterval(Poly(Y, [10, 1]), Interval(3, oo)),
    ])
    assert expr == event

def test_parse_5_lopen():
    # (2*x + 10 < 4) & (x + 10 >= 3)
    expr = ((2*X + 10) <= 4) & (X + 10 > 3)
    event = EventAnd([
        EventInterval(Poly(Y, [10, 2]), Interval(-oo, 4)),
        EventInterval(Poly(Y, [10, 1]), Interval.open(3, oo)),
    ])
    assert expr == event

def test_parse_6():
    # (x**2 - 2*x) > 10
    expr = (X**2 - 2*X) > 10
    event = EventInterval(Poly(Y, [0, -2, 1]), Interval.open(10, oo))
    assert expr == event

    # (exp(x)**2 - 2*exp(x)) > 10
    expr = (Exp(X)**2 - 2*Exp(X)) > 10
    event = EventInterval(Poly(Exp(X), [0, -2, 1]), Interval.open(10, oo))
    assert expr == event

def test_parse_7():
    # Illegal expression, cannot express in our custom DSL.
    with pytest.raises(ValueError):
        (X**2 - 2*X + Exp(X)) > 10

def test_parse_8():
    Z = Identity('Z')
    with pytest.raises(ValueError):
        (X + Z) < 3

def test_parse_9_open():
    # 2(log(x))**3 - log(x) -5 > 0
    expr = 2*(Log(X))**3 - Log(X) - 5
    expr_prime = Poly(Log(Y), [-5, -1, 0, 2])
    assert expr == expr_prime

    event = EventInterval(expr, Interval.open(0, oo))
    assert (expr > 0) == event

    # Cannot add polynomials with different subexpressions.
    with pytest.raises(ValueError):
        (2*Log(X))**3 - Log(X) - 5

def test_parse_10():
    # exp(sqrt(log(x))) > -5
    expr = Exp(Sqrt(Log(Y)))
    event = EventInterval(expr, Interval.open(-5, oo))
    assert (expr > -5) == event

def test_parse_11_open():
    # exp(sqrt(log(x))) > 6
    pass

def test_parse_12():
    # 2*sqrt(|x|) - 3 > 10
    expr = 2*Sqrt(Abs(X)) - 3
    expr_prime = Poly(Sqrt(Abs(Y)), [-3, 2])
    assert expr == expr_prime

    event = EventInterval(expr, Interval.open(10, oo))
    assert (expr > 10) == event

def test_parse_13():
    # 2*sqrt(|x|**2) - 3 > 10
    expr = 2*Sqrt(Abs(X)**2) - 3
    expr_prime = Poly(Sqrt(Pow(Abs(Y), 2)), [-3, 2])
    assert expr == expr_prime

def test_parse_14():
    # x**2 > 10
    expr = X**2 > 10
    event = EventInterval(Pow(Y, 2), Interval.open(10, oo))
    assert expr == event

def test_parse_15():
    # ((x**4)**(1/7)) < 9
    expr = ((X**4))**Rat(1, 7)
    expr_prime = Radical(Pow(Y, 4), 7)
    assert expr == expr_prime

    event = EventInterval(expr_prime, Interval.open(-oo, 9))
    assert (expr < 9) == event

def test_parse_16():
    # (x**(1/7))**4 < 9
    for expr in [((X**Rat(1,7)))**4, (X**(1,7))**4]:
        expr_prime = Pow(Radical(Y, 7), 4)
        assert expr == expr_prime

    event = EventInterval(expr, Interval.open(-oo, 9))
    assert (expr < 9) == event

def test_parse_17():
    # https://www.wolframalpha.com/input/?i=Expand%5B%2810%2F7+%2B+X%29+%28-1%2F%285+Sqrt%5B2%5D%29+%2B+X%29+%28-Sqrt%5B5%5D+%2B+X%29%5D
    for Z in [X, Log(X), Abs(1+X**2)]:
        expr = (Z - Rat(1, 10) * sympy.sqrt(2)) \
            * (Z + Rat(10, 7)) \
            * (Z - sympy.sqrt(5))
        coeffs = [
            sympy.sqrt(10)/7,
            1/sympy.sqrt(10) - (10*sympy.sqrt(5))/7 - sympy.sqrt(2)/7,
            (-sympy.sqrt(5) - 1/(5 * sympy.sqrt(2))) + Rat(10)/7,
            1,
        ]
        expr_prime = Poly(Z, coeffs)
        assert expr == expr_prime

def test_parse_18():
    # 3*(x**(1/7))**4 - 3*(x**(1/7))**2 <= 9
    Z = X**Rat(1, 7)
    expr = 3*Z**4 - 3*Z**2
    expr_prime = Poly(Radical(Y, 7), [0, 0, -3, 0, 3])
    assert expr == expr_prime

    event = EventInterval(expr_prime, Interval(-oo, 9))
    assert (expr <= 9) == event

    event_not = EventInterval(expr_prime, Interval.open(9, oo))
    assert ~(expr <= 9) == event_not

    expr = (3*Abs(Z))**4 - (3*Abs(Z))**2
    expr_prime = Poly(Poly(Abs(Z), [0, 3]), [0, 0, -1, 0, 3])

def test_parse_19():
    # 3*(x**(1/7))**4 - 3*(x**(1/7))**2 <= 9
    #   or || 3*(x**(1/7))**4 - 3*(x**(1/7))**2 > 11
    Z = X**Rat(1, 7)
    expr = 3*Z**4 - 3*Z**2

    event = (expr <= 9) | (expr > 11)
    event_prime = EventOr([
        EventInterval(expr, Interval(-oo, 9)),
        EventInterval(expr, Interval.open(11, oo)),
    ])
    assert event == event_prime

    event = ((expr <= 9) | (expr > 11)) & (~(expr < 10))
    event_prime = EventAnd([
        EventOr([
            EventInterval(expr, Interval(-oo, 9)),
            EventInterval(expr, Interval.open(11, oo))]),
        EventInterval(expr, Interval(10, oo))
    ])
    assert event == event_prime

def test_parse_20():
    # log(x**2 - 3) < 5
    expr = Log(X**2 - 3)
    expr_prime = Log(Poly(Y, [-3, 0, 1]))
    event = EventInterval(expr_prime, Interval.open(-oo, 5))
    assert (expr < 5) == event

def test_parse_21__ci_():
    # 1 <= log(x**3 - 3*x + 3) < 5
    # Can only be solved by numerical approximation of roots.
    # https://www.wolframalpha.com/input/?i=1+%3C%3D+log%28x**3+-+3x+%2B+3%29+%3C+5
    expr = Log(X**3 - 3*X + 3)
    expr_prime = Log(Poly(Y, [3, -3, 0, 1]))
    assert expr == expr_prime
    assert ((1 <= expr) & (expr < 5)) \
        == EventInterval(expr, Interval.Ropen(1, 5))

def test_parse_24_negative_power():
    assert X**(-3) == Reciprocal(Pow(X, 3))
    assert X**((-3, 1)) == Reciprocal(Pow(X, 3))
    assert X**(-Rat(1, 3)) == Reciprocal(Radical(X, 3))
    with pytest.raises(ValueError):
        X**0

def test_parse_26_piecewise_one_expr_basic_event():
    assert (Y**2)*(0 <= Y) == Piecewise(
        [Poly(Y, [0, 0, 1])],
        [EventInterval(Y, Interval(0, oo))])
    assert (0 <= Y)*(Y**2) == Piecewise(
        [Poly(Y, [0, 0, 1])],
        [EventInterval(Y, Interval(0, oo))])
    assert ((0 <= Y) < 5)*(Y < 1) == Piecewise(
        [EventInterval(Y, Interval.Ropen(-oo, 1))],
        [EventInterval(Y, Interval.Ropen(0, 5))],
    )
    assert ((0 <= Y) < 5)*(~(Y < 1)) == Piecewise(
        [EventInterval(Y, Interval(1, oo))],
        [EventInterval(Y, Interval.Ropen(0, 5))],
    )
    assert 10*(0 <= Y) == Poly(
        EventInterval(Y, Interval(0, oo)),
        [0, 10])

def test_parse_26_piecewise_one_expr_compound_event():
    assert (Y**2)*((Y < 0) | (0 < Y)) == Piecewise(
        [Poly(Y, [0, 0, 1])],
        [EventOr([
            EventInterval(Y, Interval.open(-oo, 0)),
            EventInterval(Y, Interval.open(0, oo)),
            ])])

    assert (Y**2)*(~((3 < Y) <= 4)) == Piecewise(
        [Poly(Y, [0, 0, 1])],
        [EventOr([
            EventInterval(Y, Interval(-oo, 3)),
            EventInterval(Y, Interval.open(4, oo)),
            ])])

def test_parse_27_piecewise_many():
    assert (Y < 0)*(Y**2) + (0 <= Y)*Y**((1, 2)) == Piecewise(
        [
            Poly(Y, [0, 0, 1]),
            Radical(Y, 2)],
        [
            EventInterval(Y, Interval.open(-oo, 0)),
            EventInterval(Y, Interval(0, oo))
        ])

def test_errors():
    with pytest.raises(ValueError):
        1 + Log(X) - Exp(X)
    with pytest.raises(TypeError):
        Log(X) ** Exp(X)
    with pytest.raises(ValueError):
        Abs(X) ** sympy.sqrt(10)
    with pytest.raises(ValueError):
        Log(X) * X
    with pytest.raises(ValueError):
        (2*Log(X)) - Rat(1, 10) * Abs(X)
    with pytest.raises(ValueError):
        X**(1.71)
    with pytest.raises(ValueError):
        Abs(X)**(1.1, 8)
    with pytest.raises(ValueError):
        Abs(X)**(7, 8)
    with pytest.raises(ValueError):
        (-3)**X
    with pytest.raises(ValueError):
        (Identity('Z')**2)*(Y > 0)
    with pytest.raises(ValueError):
        (Y > 0) * (Identity('Z')**2)
    with pytest.raises(ValueError):
        ((Y > 0) | (Identity('Z') < 3)) * (Identity('Z')**2)
    with pytest.raises(ValueError):
        Y**2 + (0 <= Y) * Y
    with pytest.raises(ValueError):
        (Y <= 0)*(Y**2) + (0 <= Y)*Y**((1, 2))
    with pytest.raises(ValueError):
        (Y <= 0)*(Y**2) + (0 <= Identity('Z'))*Y**((1, 2))
    with pytest.raises(ValueError):
        (Y <= 0)*(Y**2) + (0 <= Identity('Z'))*Identity('Z')**((1, 2))

    # TypeErrors from 'return NotImplemented'.
    with pytest.raises(TypeError):
        X + 'a'
    with pytest.raises(TypeError):
        X * 'a'
    with pytest.raises(TypeError):
        X / 'a'
    with pytest.raises(TypeError):
        X**'s'

def test_add_polynomials_power_one():
    # TODO: Update parser to handle this edge case.
    Z = X**5
    with pytest.raises(ValueError):
        # GOTCHA: The expression Z**2 is of type Poly(subexpr=Poly)
        # But Z if of type Poly(subexpr=Id)
        Z**2 + Z
    # Raise Z to power one to embed in a polynomial.
    Z**2 + Z**1

def test_negate_polynomial():
    assert -(X**2 + 2) == Poly(X, [-2, 0, -1])

def test_divide_multiplication():
    assert (X**2 + 2) / 2 == Poly(X, [1, 0, Rat(1, 2)])

def test_rdivide_reciprocal():
    assert 1 / X == Reciprocal(X)
    assert 1 / abs(X) == Reciprocal(abs(X))
    assert 3 / abs(X) == Poly(Reciprocal(abs(X)), [0, 3])
    assert 3 / abs(X) == Poly(Reciprocal(abs(X)), [0, 3])
    assert 1 / (X**2 + 2) == Reciprocal(Poly(X, [2, 0, 1]))
    assert 2 / (X**2 + 2) == Poly(Reciprocal(Poly(X, [2, 0, 1])), [0, 2])
    assert -2 / (X**2 + 2) == Poly(Reciprocal(Poly(X, [2, 0, 1])), [0, -2])
    assert 1 + 3 * (1/(abs(X))) + 10 * (1/abs(X))**2 \
         == Poly(Reciprocal(abs(X)), [1, 3, 10])

def test_event_negation_de_morgan():
    A, B, C = (X**3 < 10), (X < 1), ((3*X**2-100) << [10, -10])

    expr_a = ~(A & (B | C))
    expr_b = (~A | (~B & ~C))
    assert expr_a == expr_b

    expr_a = ~(A | (B & C))
    expr_b = (~A & (~B | ~C))
    assert expr_a == expr_b

def test_event_sequential_parse():
    A, B, C , D = (X > 10), (X**2 < 5), (X**3 < 0), (2*X < 7)
    assert A & B & ~C == EventAnd([A, B, ~C])
    assert A & (B & ~C) == EventAnd([A, B, ~C])
    assert (A & B) & (~C & D) == EventAnd([A, B, ~C, D])

    assert A | B | ~C == EventOr([A, B, ~C])
    assert A | (B | ~C) == EventOr([A, B, ~C])
    assert (A | B) | (~C | D) == EventOr([A, B, ~C, D])

    assert A | B | (~C & D) == EventOr([A, B, ~C & D])
    assert A & B & (~C | D) == EventAnd([A, B, ~C | D])

    assert (~C & D) | A | B == EventOr([~C & D, A, B])
    assert (~C & D) | (A | B) == EventOr([~C & D, A, B])

    assert (~C | D) & A & B == EventAnd([~C | D, A, B])
    assert (~C | D) & (A & B) == EventAnd([~C | D, A, B])

def test_event_inequality_parse():
    assert (5 < (X < 10)) \
        == ((5 < X) < 10) \
        == EventInterval(X, Interval.open(5, 10))
    assert (5 <= (X < 10)) \
        == ((5 <= X) < 10) \
        == EventInterval(X, Interval.Ropen(5, 10))
    assert (5 < (X <= 10)) \
        == ((5 < X) <= 10) \
        == EventInterval(X, Interval.Lopen(5, 10))
    assert (5 <= (X <= 10)) \
        == ((5 <= X) <= 10) \
        == EventInterval(X, Interval(5, 10))
    # GOTCHA: This expression is syntactically equivalent to
    # (5 < X) and (X < 10)
    # Since and short circuits and 5 < X is not False,
    # return value of expression is X < 10
    assert (5 < X < 10) == (X < 10)

    # Yields a finite set .
    assert ((5 < X) < 5) == EventFiniteReal(X, EmptySet)
    assert ((5 < X) <= 5) == EventFiniteReal(X, EmptySet)
    assert ((5 <= X) < 5) == EventFiniteReal(X, EmptySet)
    assert ((5 <= X) <= 5) == EventFiniteReal(X, FiniteReal(5))

    # Negated single interval.
    assert ~(5 < X) == (X <= 5)
    assert ~(5 <= X) == (X < 5)
    assert ~(X < 5) == (5 <= X)
    assert ~(X <= 5) == (5 < X)

    # Negated union of two intervals.
    assert ~(5 < (X < 10)) \
        == ~((5 < X) < 10) \
        == (X <= 5) | (10 <= X)
    assert ~(5 <= (X < 10)) \
        == ~((5 <= X) < 10) \
        == (X < 5) | (10 <= X)
    assert ~(5 < (X <= 10)) \
        == ~((5 < X) <= 10) \
        == (X <= 5) | (10 < X)
    assert ~(5 <= (X <= 10)) \
        == ~((5 <= X) <= 10) \
        == (X < 5) | (10 < X)
    assert ~((10 < X) < 5) \
        == ~(10 < (X < 5)) \
        == EventInterval(X, Interval(-oo, oo))

    # A complicated negated union.
    assert ((~(X < 5)) < 10) \
        == ((5 <= X) < 10) \
        == EventInterval(X, Interval.Ropen(5, 10))

def test_event_inequality_parse_errors():
    # EventInterval.
    with pytest.raises(ValueError):
        (X <= 10) < 5
    with pytest.raises(ValueError):
        (X <= 10) <= 5
    with pytest.raises(ValueError):
        5 < (10 <= X)
    with pytest.raises(ValueError):
        5 <= (10 <= X)
    # EventSet.
    with pytest.raises(TypeError):
        5 < (X << (10, 12))
    with pytest.raises(TypeError):
        (X << (10, 12)) < 5
    # Complement., since ~(X < 10) = (10 <= X)
    with pytest.raises(ValueError):
        5 < (~(X < 10))
    # Type mismatch.
    with pytest.raises(TypeError):
        X < 'a'
    with pytest.raises(TypeError):
        X <= 'a'
    with pytest.raises(TypeError):
        X > 'a'
    with pytest.raises(TypeError):
        X >= 'a'

def test_event_inequality_string():
    assert str(5 < X) == '5 < X'
    assert str(5 <= X) == '5 <= X'
    assert str(X < 10) == str(10 > X) == 'X < 10'
    assert str(X <= 10) == str(10 >= X) == 'X <= 10'
    assert str(5 < (X < 10)) == '(5 < X) < 10'
    assert str(5 <= (X < 10)) == '(5 <= X) < 10'
    assert str(5 < (X <= 10)) == '(5 < X) <= 10'
    assert str(5 <= (X <= 10)) == '(5 <= X) <= 10'
    assert str((X < 10) & (X < 5)) == 'X < 5'
    assert str((X < 10) | (X < 5)) == 'X < 10'

def test_event_containment_string():
    assert str(X << [10, 1]) == 'X << {1, 10}'
    assert str(X << {1, 2}) == 'X << {1, 2}'
    assert str(X << FiniteReal(1, 11)) == 'X << {1, 11}'

def test_event_containment_real():
    assert (X << Interval(0, 10)) == EventInterval(X, Interval(0, 10))
    for values in [FiniteReal(0, 10), [0, 10], {0, 10}]:
        assert (X << values) == EventFiniteReal(X, FiniteReal(0, 10))
    # with pytest.raises(ValueError):
    #     X << {1, None}
    assert X << {1, 2} == EventFiniteReal(X, {1, 2})
    assert ~(X << {1, 2}) == EventOr([
        EventInterval(X, Interval.Ropen(-oo, 1)),
        EventInterval(X, Interval.open(1, 2)),
        EventInterval(X, Interval.Lopen(2, oo)),
    ])
    # https://github.com/probcomp/sum-product-dsl/issues/22
    # and of EventBasic does not yet perform simplifications.
    assert ~(~(X << {1, 2})) == \
        ((1 <= X) & ((X <= 1) | (2 <= X)) & (X <= 2))

def test_event_containment_nominal():
    assert X << {'a'} == EventFiniteNominal(X, FiniteNominal('a'))
    assert ~(X << {'a'}) == EventFiniteNominal(X, FiniteNominal('a',b=True))
    assert ~(~(X << {'a'})) == X << {'a'}

def test_event_containment_mixed():
    assert X << {1, 2, 'a'} \
        == (X << {1,2}) | (X << {'a'}) \
        == EventOr([
        EventFiniteReal(X, {1, 2}),
        EventFiniteNominal(X, FiniteNominal('a'))
    ])
    assert (~(X << {1, 2, 'a'})) == ~(X << {1,2}) & ~(X << {'a'})

    # https://github.com/probcomp/sum-product-dsl/issues/22
    # Taking the And of EventBasic does not perform simplifications.
    assert ~(~(X << {1, 2, 'a'})) == EventOr([
        ((1 <= X) & ((X <= 1) | (2 <= X)) & (X <= 2)),
        X << {'a'}
    ])

def test_event_containment_union():
    assert (X << (Interval(0, 1) | Interval(2, 3))) \
        == (((0 <= X) <= 1) | ((2 <= X) <= 3))
    assert (X << (FiniteReal(0, 1) | Interval(2, 3))) \
        == ((X << {0, 1}) | ((2 <= X) <= 3))
    assert (X << FiniteNominal('a', b=True)) \
        == EventFiniteNominal(X, FiniteNominal('a', b=True))
    assert X << EmptySet == EventFiniteReal(X, EmptySet)
    # Ordering is not guaranteed.
    a = X << (Interval(0,1) | (FiniteReal(1.5) | FiniteNominal('a')))
    assert len(a.subexprs) == 3
    assert EventInterval(X, Interval(0,1)) in a.subexprs
    assert EventFiniteReal(X, FiniteReal(1.5)) in a.subexprs
    assert EventFiniteNominal(X, FiniteNominal('a')) in a.subexprs

def test_event_basic_simplifications():
    assert (1 < X) & (3 < X) == (3 < X)
    assert (1 < X) & (X < 5) == ((1 < X) < 5)
    assert (X < 1) & (X >= -3) == ((-3 <= X) < 1)
    assert (X >= -3) & (X << {1}) == X << {1}
    assert (X << {1, -10}) & (X >= -3) == X << {1}
    assert (X << {1, -10}) & (X << {-3}) == (X << EmptySet)
    assert (X << {1, -3}) & (X << {-3}) == X << {-3}

    assert (1 < X) | (3 < X) == (1 < X)
    assert (1 < X) | (X << {2}) == (1 < X)
    assert (X << {7}) | (1 < X)  == (1 < X)
    assert (X << {-10}) | (1 < X) == EventOr([X<<{-10}, 1<X])
    assert (X << {2}) | (X << {3,4}) == (X << {2, 3, 4})
    assert (X << {'2'}) | (X << {'a'}) == X << {'2', 'a'}
    assert (X << {'2'}) & (~(X << {'a'})) == (X << {'2'})
    assert (X << {1}) & (~(X << {'a'})) == (X << EmptySet)

    assert ~(X << {2}) & (X << {'a'}) == EventAnd([~(X << {2}), X <<{'a'}])
    assert (X < 1) & (X << {'a'}) == X << EmptySet
    assert ((0 <= X) < 1) & (~(X << {'a'})) == X << EmptySet

    assert (X << {2}) & (X << {'a'}) == X << EmptySet
    assert (X << {'a'}) & (X << {1}) == X << EmptySet
    assert (X << EmptySet) & (X << EmptySet) == X << EmptySet
    assert (X << EmptySet) & (~(X << EmptySet)) == X << EmptySet
    assert (X << {'a'}) & ~(X << {'b'}) == (X << {'a'})

    # Complement case in Event.__or__.
    assert (X << {1}) | (~(X << {'a'})) == EventOr([X<<{1}, ~(X<<{'a'})])
    assert (X << {'a'}) | ~(X << {'b'}) == ~(X << {'b'})

def test_event_complex_simplification():
    # De Morgan's case in EventFinteNominal.__or__.
    assert ~(X << {'a'}) | ~(X << {'b'}) == EventFiniteNominal(X, FiniteNominal(b=True))
    assert ~(X << {'a'}) | ~(X << {'a', 'b'}) == ~(X << {'a'})
    assert (X << {'1'}) | ~(X << {'1', '2'}) == ~(X << {'2'})

def test_xor():
    A = X << {'a'}
    B = ~(X << {'b'})
    assert (A ^ B) \
        == ((A & ~B) | (~A & B)) \
        == EventFiniteNominal(X, FiniteNominal('a', 'b', b=True))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from sympy import Poly as SymPoly
from sympy import Rational
from sympy import sqrt as SymSqrt
from sympy.abc import x

from sppl.poly import solve_poly_equality
from sppl.poly import solve_poly_inequality

from sppl.math_util import allclose

from sppl.sets import EmptySet
from sppl.sets import ExtReals
from sppl.sets import FiniteReal
from sppl.sets import Interval
from sppl.sets import Reals
from sppl.sets import Union
from sppl.sets import inf as oo

def test_solve_poly_inequaltiy_pos_inf():
    assert solve_poly_inequality(x**2-10*x+100, oo, True) == Reals
    assert solve_poly_inequality(x**2-10*x+100, oo, False) == ExtReals

    assert solve_poly_inequality(-x**3+10*x, oo, False) == ExtReals
    assert solve_poly_inequality(-x**3+10*x, oo, True) == Reals | FiniteReal(oo)

    assert solve_poly_inequality(x**3-10*x, oo, False) == ExtReals
    assert solve_poly_inequality(x**3-10*x, oo, True) == Reals | FiniteReal(-oo)


def test_solve_poly_inequaltiy_neg_inf():
    assert solve_poly_inequality(x**2-10*x+100, -oo, True) is EmptySet
    assert solve_poly_inequality(x**2-10*x+100, -oo, False) is EmptySet

    assert solve_poly_inequality(x**3-10*x, -oo, True) is EmptySet
    assert solve_poly_inequality(x**3-10*x, -oo, False) == FiniteReal(-oo)

    assert solve_poly_inequality(-x**2+10*x+100, -oo, True) is EmptySet
    assert solve_poly_inequality(-x**2+10*x+100, -oo, False) == FiniteReal(-oo, oo)


p_quadratic = SymPoly((x-SymSqrt(2)/10)*(x+Rational(10, 7)), x)
expr_quadratic = p_quadratic.args[0]
def test_solve_poly_equality_quadratic_zero():
    roots = solve_poly_equality(expr_quadratic, 0)
    assert roots == FiniteReal(SymSqrt(2)/10, -Rational(10,7))
def test_solve_poly_inequality_quadratic_zero():
    interval = solve_poly_inequality(expr_quadratic, 0, False)
    assert interval == Interval(-Rational(10,7), SymSqrt(2)/10)
    interval = solve_poly_inequality(expr_quadratic, 0, True)
    assert interval == Interval.open(-Rational(10,7), SymSqrt(2)/10)

xe1_quad0 = -5/7 + SymSqrt(2)/20 + SymSqrt(2)*SymSqrt(700*SymSqrt(2) + 14849)/140
xe1_quad1 = -SymSqrt(2)*SymSqrt(700*SymSqrt(2) + 14849)/140 - 5/7 + SymSqrt(2)/20
def test_solve_poly_equality_quadratic_one():
    roots = solve_poly_equality(expr_quadratic, 1)
    # SymPy is not smart enough to simplify irrational roots symbolically
    # so check numerical equality of the symbolic roots.
    assert len(roots) == 2
    assert any(allclose(float(x), float(xe1_quad0)) for x in roots)
    assert any(allclose(float(x), float(xe1_quad1)) for x in roots)


p_cubic_int = SymPoly((x-1)*(x+2)*(x-11), x)
expr_cubic_int = p_cubic_int.args[0]
def test_solve_poly_equality_cubic_int_zero():
    roots = solve_poly_equality(expr_cubic_int, 0)
    assert roots == FiniteReal(-2, 1, 11)
def test_solve_poly_inequality_cubic_int_zero():
    interval = solve_poly_inequality(expr_cubic_int, 0, False)
    assert interval == Interval(-oo, -2) | Interval(1, 11)
    interval = solve_poly_inequality(expr_cubic_int, 0, True)
    assert interval == Interval.open(-oo, -2) | Interval.open(1, 11)

xe1_cubic_int0 = -1.97408387376586
xe1_cubic_int1 = 0.966402009973818
xe1_cubic_int2 = 11.007681863792
def test_solve_poly_equality_cubic_int_one():
    roots = solve_poly_equality(expr_cubic_int, 1)
    assert len(roots) == 3
    assert any(allclose(float(x), xe1_cubic_int0) for x in roots)
    assert any(allclose(float(x), xe1_cubic_int1) for x in roots)
    assert any(allclose(float(x), xe1_cubic_int2) for x in roots)
def test_solve_poly_inequality_cubic_int_one():
    interval = solve_poly_inequality(expr_cubic_int, 1, True)
    assert isinstance(interval, Union)
    assert len(interval.args)==2
    # First interval.
    assert interval.args[0].left == -oo
    assert allclose(float(interval.args[0].right), xe1_cubic_int0)
    assert interval.args[0].right_open
    # Second interval.
    assert allclose(float(interval.args[1].left), xe1_cubic_int1)
    assert allclose(float(interval.args[1].right), xe1_cubic_int2)
    assert interval.args[1].left_open
    assert interval.args[1].right_open

    interval = solve_poly_inequality(-1*expr_cubic_int, -1, True)
    assert isinstance(interval, Union)
    assert len(interval.args) == 2
    # First interval.
    assert allclose(float(interval.args[0].left), xe1_cubic_int0)
    assert allclose(float(interval.args[0].right), xe1_cubic_int1)
    # Second interval.
    assert allclose(float(interval.args[1].left), xe1_cubic_int2)
    assert interval.args[1].right == oo


p_cubic_irrat = SymPoly((x-SymSqrt(2)/10)*(x+Rational(10, 7))*(x-SymSqrt(5)), x)
expr_cubic_irrat = p_cubic_irrat.args[0]
def test_solve_poly_equality_cubic_irrat_zero():
    roots = solve_poly_equality(expr_cubic_irrat, 0)
    # Confirm that roots contains symbolic elements (no timeout).
    assert -Rational(10,7) in roots
    # SymPy is not smart enough to simplify irrational roots symbolically
    # so check numerical equality of the symbolic roots.
    assert any(allclose(float(x), float(SymSqrt(2)/10)) for x in roots)
    assert any(allclose(float(x), float(SymSqrt(5))) for x in roots)

xe1_cubic_irrat0 = -1.21493150246058
xe1_cubic_irrat1 = -0.19158140952462
xe1_cubic_irrat2 = 2.35543081715088
def test_solve_poly_equality_cubic_irrat_one():
    # This expression is too slow to solve symbolically.
    # The 5s timeout will trigger a numerical approximation.
    roots = solve_poly_equality(expr_cubic_irrat, 1)
    assert len(roots) == 3
    assert any(allclose(float(x), xe1_cubic_irrat0) for x in roots)
    assert any(allclose(float(x), xe1_cubic_irrat1) for x in roots)
    assert any(allclose(float(x), xe1_cubic_irrat2) for x in roots)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from sppl.compilers.spn_to_sppl import render_sppl
from sppl.compilers.sppl_to_python import SPPL_Compiler
from sppl.math_util import allclose
from sppl.tests.test_render import get_model

def test_render_sppl():
    model = get_model()
    sppl_code = render_sppl(model)
    compiler = SPPL_Compiler(sppl_code.getvalue())
    namespace = compiler.execute_module()
    (X, Y) = (namespace.X, namespace.Y)
    for i in range(5):
        assert allclose(model.logprob(Y << {'0'}), [
            model.logprob(Y << {str(i)}),
            namespace.model.logprob(Y << {str(i)})
        ])
    for i in range(4):
        assert allclose(
            model.logprob(X << {i}),
            namespace.model.logprob(X << {i}))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

import numpy

from sppl.distributions import poisson
from sppl.distributions import randint
from sppl.math_util import allclose
from sppl.math_util import logdiffexp
from sppl.math_util import logsumexp
from sppl.sets import Interval
from sppl.sets import Range
from sppl.sets import inf as oo
from sppl.spn import DiscreteLeaf
from sppl.spn import SumSPN
from sppl.transforms import Id

def test_poisson():
    X = Id('X')
    spn = X >> poisson(mu=5)

    a = spn.logprob((1 <= X) <= 7)
    b = spn.logprob(X << {1,2,3,4,5,6,7})
    c = logsumexp([spn.logprob(X << {i}) for i in range(1, 8)])
    assert allclose(a, b)
    assert allclose(a, c)
    assert allclose(b, c)

    spn_condition = spn.condition(10 <= X)
    assert spn_condition.conditioned
    assert spn_condition.support == Range(10, oo)
    assert spn_condition.logZ == logdiffexp(0, spn.logprob(X<=9))

    assert allclose(
        spn_condition.logprob(X <= 10),
        spn_condition.logprob(X << {10}))
    assert allclose(
        spn_condition.logprob(X <= 10),
        spn_condition.logpdf({X: 10}))

    samples = spn_condition.sample(100)
    assert all(10 <= s[X] for s in samples)

    # Unify X = 5 with left interval to make one distribution.
    event = ((1 <= X) < 5) | ((3*X + 1) << {16})
    spn_condition = spn.condition(event)
    assert isinstance(spn_condition, DiscreteLeaf)
    assert spn_condition.conditioned
    assert spn_condition.xl == 1
    assert spn_condition.xu == 5
    assert spn_condition.support == Range(1, 5)
    samples = spn_condition.sample(100, prng=numpy.random.RandomState(1))
    assert all(event.evaluate(s) for s in samples)

    # Ignore X = 14/3 as a probability zero condition.
    spn_condition = spn.condition(((1 <= X) < 5) | (3*X + 1) << {15})
    assert isinstance(spn_condition, DiscreteLeaf)
    assert spn_condition.conditioned
    assert spn_condition.xl == 1
    assert spn_condition.xu == 4
    assert spn_condition.support == Interval.Ropen(1,5)

    # Make a mixture of two components.
    spn_condition = spn.condition(((1 <= X) < 5) | (3*X + 1) << {22})
    assert isinstance(spn_condition, SumSPN)
    xl = spn_condition.children[0].xl
    idx0 = 0 if xl == 7 else 1
    idx1 = 1 if xl == 7 else 0
    assert spn_condition.children[idx1].conditioned
    assert spn_condition.children[idx1].xl == 1
    assert spn_condition.children[idx1].xu == 4
    assert spn_condition.children[idx0].conditioned
    assert spn_condition.children[idx0].xl == 7
    assert spn_condition.children[idx0].xu == 7
    assert spn_condition.children[idx0].support == Range(7, 7)

    # Condition on probability zero event.
    with pytest.raises(ValueError):
        spn.condition(((-3 <= X) < 0) | (3*X + 1) << {20})

    # Condition on FiniteReal contiguous.
    spn_condition = spn.condition(X << {1,2,3})
    assert spn_condition.xl == 1
    assert spn_condition.xu == 3
    assert allclose(spn_condition.logprob((1 <= X) <=3), 0)

    # Condition on single point.
    assert allclose(0, spn.condition(X << {2}).logprob(X<<{2}))

def test_condition_non_contiguous():
    X = Id('X')
    spn = X >> poisson(mu=5)
    # FiniteSet.
    for c in [{0,2,3}, {-1,0,2,3}, {-1,0,2,3,'z'}]:
        spn_condition = spn.condition((X << c))
        assert isinstance(spn_condition, SumSPN)
        assert allclose(0, spn_condition.children[0].logprob(X<<{0}))
        assert allclose(0, spn_condition.children[1].logprob(X<<{2,3}))
    # FiniteSet or Interval.
    spn_condition = spn.condition((X << {-1,'x',0,2,3}) | (X > 7))
    assert isinstance(spn_condition, SumSPN)
    assert len(spn_condition.children) == 3
    assert allclose(0, spn_condition.children[0].logprob(X<<{0}))
    assert allclose(0, spn_condition.children[1].logprob(X<<{2,3}))
    assert allclose(0, spn_condition.children[2].logprob(X>7))

def test_randint():
    X = Id('X')
    spn = X >> randint(low=0, high=5)
    assert spn.xl == 0
    assert spn.xu == 4
    assert spn.logprob(X < 5) == spn.logprob(X <= 4) == 0
    # i.e., X is not in [0, 3]
    spn_condition = spn.condition(~((X+1) << {1, 4}))
    assert isinstance(spn_condition, SumSPN)
    xl = spn_condition.children[0].xl
    idx0 = 0 if xl == 1 else 1
    idx1 = 1 if xl == 1 else 0
    assert spn_condition.children[idx0].xl == 1
    assert spn_condition.children[idx0].xu == 2
    assert spn_condition.children[idx1].xl == 4
    assert spn_condition.children[idx1].xu == 4
    assert allclose(spn_condition.children[idx0].logprob(X<<{1,2}), 0)
    assert allclose(spn_condition.children[idx1].logprob(X<<{4}), 0)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from collections import OrderedDict
from itertools import chain
from itertools import combinations
from math import isinf

import numpy
import sympy

from sympy.core.relational import Relational

from .sets import FiniteReal
from .sets import Interval

def get_symbols(expr):
    atoms = expr.atoms()
    return [a for a in atoms if isinstance(a, sympy.Symbol)]

def get_union(sets):
    return sets[0].union(*sets[1:])

def get_intersection(sets):
    return sets[0].intersection(*sets[1:])

def are_disjoint(sets):
    union = get_union(sets)
    return len(union) == sum(len(s) for s in sets)

def are_identical(sets):
    intersection = get_intersection(sets)
    return all(len(s) == len(intersection) for s in sets)

def binspace(start, stop, num=10):
    values = numpy.linspace(start, stop, num)
    bins = list(zip(values[:-1], values[1:]))
    return [Interval(*b) for b in bins]

def powerset(values, start=0):
    s = list(values)
    subsets = (combinations(s, k) for k in range(start, len(s) + 1))
    return chain.from_iterable(subsets)

def partition_list_blocks(values):
    partition = OrderedDict([])
    for i, v in enumerate(values):
        x = hash(v)
        if x not in partition:
            partition[x] = []
        partition[x].append(i)
    return list(partition.values())

def partition_finite_real_contiguous(x):
    # Convert FiniteReal to list of FiniteReal, each with contiguous values.
    assert isinstance(x, FiniteReal)
    values = sorted(x.values)
    blocks = [[values[0]]]
    for y in values[1:]:
        expected = blocks[-1][-1] + 1
        if y == expected:
            blocks[-1].append(y)
        else:
            blocks.append([y])
    return [FiniteReal(*v) for v in blocks]

def sympify_number(x):
    if isinstance(x, (int, float)):
        return x
    msg = 'Expected a numeric term, not %s' % (x,)
    try:
        # String fallback in sympify has been deprecated since SymPy 1.6. Use
        # sympify(str(obj)) or sympy.core.sympify.converter or obj._sympy_
        # instead. See https://github.com/sympy/sympy/issues/18066 for more
        # info.
        sym = sympy.sympify(str(x))
        if not sym.is_number:
            raise TypeError(msg)
        return sym
    except (sympy.SympifyError, AttributeError, TypeError):
        raise TypeError(msg)

def sym_log(x):
    assert 0 <= x
    if x == 0:
        return -float('inf')
    if isinf(x):
        return float('inf')
    return sympy.log(x)

def sympy_solver(expr):
    # Sympy is buggy and slow.  Use Transforms.
    symbols = get_symbols(expr)
    if len(symbols) != 1:
        raise ValueError('Expression "%s" needs exactly one symbol.' % (expr,))

    if isinstance(expr, Relational):
        result = sympy.solveset(expr, domain=sympy.Reals)
    elif isinstance(expr, sympy.Or):
        subexprs = expr.args
        intervals = [sympy_solver(e) for e in subexprs]
        result = sympy.Union(*intervals)
    elif isinstance(expr, sympy.And):
        subexprs = expr.args
        intervals = [sympy_solver(e) for e in subexprs]
        result = sympy.Intersection(*intervals)
    elif isinstance(expr, sympy.Not):
        (notexpr,) = expr.args
        interval = sympy_solver(notexpr)
        result = interval.complement(sympy.Reals)
    else:
        raise ValueError('Expression "%s" has unknown type.' % (expr,))

    if isinstance(result, sympy.ConditionSet):
        raise ValueError('Expression "%s" is not invertible.' % (expr,))

    return result



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from fractions import Fraction
from math import log

import pytest

import numpy

from sppl.distributions import choice
from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.math_util import logsumexp
from sppl.sets import FiniteNominal
from sppl.sets import Interval
from sppl.sets import inf as oo
from sppl.spn import ContinuousLeaf
from sppl.spn import ExposedSumSPN
from sppl.spn import NominalLeaf
from sppl.spn import ProductSPN
from sppl.spn import SumSPN
from sppl.transforms import Id

def test_sum_normal_gamma():
    X = Id('X')
    weights = [
        log(Fraction(2, 3)),
        log(Fraction(1, 3))
    ]
    spn = SumSPN(
        [X >> norm(loc=0, scale=1), X >> gamma(loc=0, a=1),], weights)

    assert spn.logprob(X > 0) == logsumexp([
        spn.weights[0] + spn.children[0].logprob(X > 0),
        spn.weights[1] + spn.children[1].logprob(X > 0),
    ])
    assert spn.logprob(X < 0) == log(Fraction(2, 3)) + log(Fraction(1, 2))
    samples = spn.sample(100, prng=numpy.random.RandomState(1))
    assert all(s[X] for s in samples)
    spn.sample_func(lambda X: abs(X**3), 100)
    with pytest.raises(ValueError):
        spn.sample_func(lambda Y: abs(X**3), 100)

    spn_condition = spn.condition(X < 0)
    assert isinstance(spn_condition, ContinuousLeaf)
    assert spn_condition.conditioned
    assert spn_condition.logprob(X < 0) == 0
    samples = spn_condition.sample(100)
    assert all(s[X] < 0 for s in samples)

    assert spn.logprob(X < 0) == logsumexp([
        spn.weights[0] + spn.children[0].logprob(X < 0),
        spn.weights[1] + spn.children[1].logprob(X < 0),
    ])

def test_sum_normal_gamma_exposed():
    X = Id('X')
    W = Id('W')
    weights = W >> choice({
        '0': Fraction(2,3),
        '1': Fraction(1,3),
    })
    children = {
        '0': X >> norm(loc=0, scale=1),
        '1': X >> gamma(loc=0, a=1),
    }
    spn = ExposedSumSPN(children, weights)

    assert spn.logprob(W << {'0'}) == log(Fraction(2, 3))
    assert spn.logprob(W << {'1'}) == log(Fraction(1, 3))
    assert allclose(spn.logprob((W << {'0'}) | (W << {'1'})), 0)
    assert spn.logprob((W << {'0'}) & (W << {'1'})) == -float('inf')

    assert allclose(
        spn.logprob((W << {'0', '1'}) & (X < 1)),
        spn.logprob(X < 1))

    assert allclose(
        spn.logprob((W << {'0'}) & (X < 1)),
        spn.weights[0] + spn.children[0].logprob(X < 1))

    spn_condition = spn.condition((W << {'1'}) | (W << {'0'}))
    assert isinstance(spn_condition, SumSPN)
    assert len(spn_condition.weights) == 2
    assert \
        allclose(spn_condition.weights[0], log(Fraction(2,3))) \
            and allclose(spn_condition.weights[0], log(Fraction(2,3))) \
        or \
        allclose(spn_condition.weights[1], log(Fraction(2,3))) \
            and allclose(spn_condition.weights[0], log(Fraction(2,3))
        )

    spn_condition = spn.condition((W << {'1'}))
    assert isinstance(spn_condition, ProductSPN)
    assert isinstance(spn_condition.children[0], NominalLeaf)
    assert isinstance(spn_condition.children[1], ContinuousLeaf)
    assert spn_condition.logprob(X < 5) == spn.children[1].logprob(X < 5)

def test_sum_normal_nominal():
    X = Id('X')
    children = [
        X >> norm(loc=0, scale=1),
        X >> choice({'low': Fraction(3, 10), 'high': Fraction(7, 10)}),
    ]
    weights = [log(Fraction(4,7)), log(Fraction(3, 7))]

    spn = SumSPN(children, weights)

    assert allclose(
        spn.logprob(X < 0),
        log(Fraction(4,7)) + log(Fraction(1,2)))

    assert allclose(
        spn.logprob(X << {'low'}),
        log(Fraction(3,7)) + log(Fraction(3, 10)))

    # The semantics of ~(X<<{'low'}) are (X << String and X != 'low')
    assert allclose(
        spn.logprob(~(X << {'low'})),
        spn.logprob((X << {'high'})))
    assert allclose(
        spn.logprob((X<<FiniteNominal(b=True)) & ~(X << {'low'})),
        spn.logprob((X<<FiniteNominal(b=True)) & (X << {'high'})))

    assert isinf_neg(spn.logprob((X < 0) & (X << {'low'})))

    assert allclose(
        spn.logprob((X < 0) | (X << {'low'})),
        logsumexp([spn.logprob(X < 0), spn.logprob(X << {'low'})]))

    assert isinf_neg(spn.logprob(X << {'a'}))
    assert allclose(
        spn.logprob(~(X << {'a'})),
        spn.logprob(X<<{'low','high'}))

    assert allclose(
        spn.logprob(X**2 < 9),
        log(Fraction(4, 7)) + spn.children[0].logprob(X**2 < 9))

    spn_condition = spn.condition(X**2 < 9)
    assert isinstance(spn_condition, ContinuousLeaf)
    assert spn_condition.support == Interval.open(-3, 3)

    spn_condition = spn.condition((X**2 < 9) | X << {'low'})
    assert isinstance(spn_condition, SumSPN)
    assert spn_condition.children[0].support == Interval.open(-3, 3)
    assert spn_condition.children[1].support == FiniteNominal('low', 'high')
    assert isinf_neg(spn_condition.children[1].logprob(X << {'high'}))

    assert spn_condition == spn.condition((X**2 < 9) | ~(X << {'high'}))
    assert allclose(spn.logprob((X < oo) | ~(X << {'1'})), 0)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

"""Convert SPPL to Python 3."""

import ast
import inspect
import copy
import io
import os
import re

from types import SimpleNamespace
from collections import namedtuple
from collections import OrderedDict
from contextlib import contextmanager

from astunparse import unparse

def __load_spn_distributions():
    from .. import distributions
    members = inspect.getmembers(distributions,lambda t: isinstance(t, type))
    return frozenset(m for (m, v) in members if m[0].islower())

DISTRIBUTIONS = __load_spn_distributions()
get_indentation = lambda i: ' ' * i

@contextmanager
def maybe_sequence_block(visitor, value, first=None):
    # Enter
    active = len(value) > 1 or first
    if active:
        idt = get_indentation(visitor.indentation)
        visitor.command.write('%sSequence(' % (idt,))
        visitor.command.write('\n')
        visitor.indentation += 4
    # Yield
    yield None
    # Exit
    if active:
        visitor.indentation -= 4
        idt = get_indentation(visitor.indentation)
        visitor.command.write('%s)' % (idt,))
        if not first:
            visitor.command.write(',')
            visitor.command.write('\n')

class SPPL_Visitor(ast.NodeVisitor):
    def __init__(self, stream=None):
        self.first = True
        self.indentation = 0
        self.context = ['global']
        self.command = stream or io.StringIO()
        self.imports = []
        self.constants = OrderedDict()
        self.variables = OrderedDict()
        self.distributions = OrderedDict()

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                with maybe_sequence_block(self, value, self.first):
                    self.first = False
                    for item in value:
                        if isinstance(item, ast.AST):
                            self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def visit_Import(self, node):
        str_node = unparse(node).replace(os.linesep, '')
        self.imports.append(str_node)
    def visit_ImportFrom(self, node):
        str_node = unparse(node).replace(os.linesep, '')
        self.imports.append(str_node)

    def visit_Assign(self, node):
        str_node = unparse(node)

        # Convert IfExp to an If?
        if isinstance(node.value, ast.IfExp):
            node_prime = unroll_ifexp(node.targets, node.value)
            return self.visit(node_prime)

        # Analyze node.target.
        assert len(node.targets) == 1, unparse(node)
        assert isinstance(node.value, ast.expr), 'unknown value %s' % (str_node,)

        # Record visited distributions.
        value = node.value
        visitor_name = SPPL_Visitor_Name(DISTRIBUTIONS, self.variables)
        visitor_name.visit(value)
        for d in visitor_name.distributions:
            if d not in self.distributions:
                self.distributions[d] = None

        # Assigning an array.
        if isinstance(value, ast.Call) and value.func.id == 'array':
            return self.visit_Assign_array(node)
        # Assigning a distribution.
        if visitor_name.distributions:
            # Assigning distribution (directly).
            if isinstance(value, ast.Call) and value.func.id in DISTRIBUTIONS:
                return self.visit_Assign_sample_or_transform(node, 'Sample')
            # Assigning distribution (mixture).
            if isinstance(value, ast.BinOp) and isinstance(value.op, ast.BitOr):
                return self.visit_Assign_sample_or_transform(node, 'Sample')
            assert False, unparse(node)
        # Assigning a transform.
        if visitor_name.variables:
            return self.visit_Assign_sample_or_transform(node, 'Transform')
        # Assigning a Python variable.
        return self.visit_Assign_py(node)

    def visit_Assign_array(self, node):
        target = node.targets[0]
        unode = unparse(node)
        assert self.context == ['global'], unode         # must be global
        assert isinstance(target, ast.Name), unode       # must not be subscript
        assert target.id not in self.variables, unode    # must be fresh
        assert len(node.value.args) == 1, unode          # must be array(n)
        # assert isinstance(node.value.args[0], ast.Num) # must be num n
        # assert isinstance(node.value.args[0].n, int)   # must be int n
        # assert node.value.args[0].n > 0                # must be pos n
        n = unparse(node.value.args[0]).strip()
        self.variables[target.id] = ('array', n)

    def visit_Assign_sample_or_transform(self, node, op):
        unode = unparse(node)
        assert op in ['Sample', 'Transform'], unode
        target = node.targets[0]
        assert isinstance(target, (ast.Name, ast.Subscript)), unode
        if isinstance(target, ast.Name):
            assert 'for' not in self.context, unode
            if 'elif' not in self.context:
                assert target.id not in self.variables, unode
                self.variables[target.id] = ('variable', None)
            if 'elif' in self.context:
                assert target.id in self.variables, unode
                assert target.id in self.variables, unode
        if isinstance(target, ast.Subscript):
            assert target.value.id in self.variables, unode
            assert self.variables[target.value.id][0] == 'array', unode
        value_prime = SPPL_Transformer_Compare().visit(node.value)
        src_value = unparse(value_prime).replace(os.linesep, '')
        src_targets = unparse(node.targets).replace(os.linesep, '')
        idt = get_indentation(self.indentation)
        self.command.write('%s%s(%s, %s),' % (idt, op, src_targets, src_value))
        self.command.write('\n')

    def visit_Assign_py(self, node):
        unode = unparse(node)
        assert self.context == ['global'], unode
        if self.distributions:
            assert False, 'constants only before sampling %s' % (unode,)
        str_node = unode.strip()
        self.constants[str_node] = None

    def visit_For(self, node):
        assert isinstance(node.iter, ast.Call)
        if node.iter.func.id == 'range':
            return self.visit_For_vanilla(node)
        if node.iter.func.id == 'switch':
            return self.visit_For_switch(node)
        assert False, unparse(node)

    def visit_For_switch(self, node):
        unode = unparse(node)
        assert isinstance(node.target, (ast.Name, ast.Tuple)), unode
        assert node.iter.func.id == 'switch', unode
        assert len(node.iter.args) == 2, unode
        assert isinstance(node.iter.args[0], (ast.Name, ast.Subscript)), unode
        # Open Switch.
        self.context.append('switch')
        idt = get_indentation(self.indentation)
        symbol = unparse(node.iter.args[0]).strip()
        values = unparse(node.iter.args[1]).strip()
        translator = str.maketrans({'(':'', ')':''})
        idx = unparse(node.target).strip().translate(translator)
        self.command.write('%sSwitch(%s, %s, lambda %s:'
            % (idt, symbol, values, idx))
        self.command.write('\n')
        # Write the body.
        self.indentation += 4
        self.generic_visit(ast.Module(node.body))
        self.indentation -= 4
        # Close Switch
        idt = get_indentation(self.indentation)
        self.command.write('%s),' % (idt,))
        self.command.write('\n')
        self.context.pop()

    def visit_For_vanilla(self, node):
        unode = unparse(node)
        assert isinstance(node.target, ast.Name), unode
        assert node.iter.func.id == 'range', unode
        assert len(node.iter.args) in [1, 2], unode
        if len(node.iter.args) == 1:
            n0 = 0
            n1 = unparse(node.iter.args[0]).strip()
        if len(node.iter.args) == 2:
            n0 = unparse(node.iter.args[0]).strip()
            n1 = unparse(node.iter.args[1]).strip()
        # Open For.
        self.context.append('for')
        idt = get_indentation(self.indentation)
        idx = unparse(node.target).strip()
        self.command.write('%sFor(%s, %s, lambda %s:' % (idt, n0, n1, idx))
        self.command.write('\n')
        # Write body.
        self.indentation += 4
        self.generic_visit(ast.Module(node.body))
        self.indentation -= 4
        # Close For.
        idt = get_indentation(self.indentation)
        self.command.write('%s),' % (idt,))
        self.command.write('\n')
        self.context.pop()

    def visit_If(self, node):
        unode = unparse(node)
        unrolled = unroll_if(node)
        assert 2 <= len(unrolled), 'if needs elif/else: %s' % (unode,)
        idt = get_indentation(self.indentation)
        # Open IfElse.
        self.context.append('if')
        self.command.write('%sIfElse(' % (idt,))
        self.command.write('\n')
        # Write branches.
        self.indentation += 4
        for i, (test, body) in enumerate(unrolled):
            # Write the test.
            test_prime = SPPL_Transformer_Compare().visit(test)
            src_test = unparse(test_prime).strip()
            idt = get_indentation(self.indentation)
            self.command.write('%s%s,' % (idt, src_test))
            self.command.write('\n')
            # Write the body.
            if i > 0:
                self.context.append('elif')
            self.indentation += 4
            self.generic_visit(ast.Module(body))
            self.indentation -= 4
            if i > 0:
                self.context.pop()
        self.indentation -= 4
        # Close IfElse.
        idt = get_indentation(self.indentation)
        self.command.write('%s),' % (idt,))
        self.command.write('\n')
        self.context.pop()

    def visit_Call(self, node):
        unode = unparse(node)
        if node.func.id == 'condition':
            assert len(node.args) == 1, unode
            idt = get_indentation(self.indentation)
            event = node.args[0]
            event_prime = SPPL_Transformer_Compare().visit(event)
            src_event = unparse(event_prime).strip()
            self.command.write('%sCondition(%s),' % (idt, src_event))
            self.command.write('\n')

def unroll_if(node, current=None):
    current = [] if current is None else current
    assert isinstance(node, ast.If)
    current.append((node.test, node.body))
    # Base case, terminating at elif.
    if not node.orelse:
        return current
    # Base case, next statement is not an If.
    if not (isinstance(node.orelse[0], ast.If) and len(node.orelse)== 1):
        current.append((ast.parse('True'), node.orelse))
        return current
    # Recursive case, next statement is elif
    return unroll_if(node.orelse[0], current)

def unroll_ifexp(target, node):
    assert isinstance(node, ast.IfExp)
    expr = ast.If(node.test, ast.Assign(target, node.body), None)
    if isinstance(node.orelse, ast.IfExp):
        expr.orelse = [unroll_ifexp(target, node.orelse)]
    else:
        expr.orelse = [ast.Assign(target, node.orelse)]
    return expr

class SPPL_Transformer_Compare(ast.NodeTransformer):
    def visit_Compare(self, node):
        # TODO: Implement or/and.
        if len(node.ops) > 1:
            return node
        if isinstance(node.ops[0], ast.In):
            return ast.BinOp(
                left=node.left,
                op=ast.LShift(),
                right=node.comparators[0])
        if isinstance(node.ops[0], ast.Eq):
            return ast.BinOp(
                left=node.left,
                op=ast.LShift(),
                right=ast.Set(node.comparators))
        if isinstance(node.ops[0], ast.NotIn):
            node_copy = copy.deepcopy(node)
            node_copy.ops[0] = ast.In()
            return ast.UnaryOp(
                op=ast.Invert(),
                operand=self.visit_Compare(node_copy))
        if isinstance(node.ops[0], ast.NotEq):
            node_copy = copy.deepcopy(node)
            node_copy.ops[0] = ast.Eq()
            return ast.UnaryOp(
                op=ast.Invert(),
                operand=self.visit_Compare(node_copy))
        return node

class SPPL_Visitor_Name(ast.NodeVisitor):
    def __init__(self, distributions, variables):
        self.distributions_lookup = distributions
        self.variables_lookup = variables
        self.distributions = set()
        self.variables = set()
    def visit_Name(self, node):
        if node.id in self.distributions_lookup:
            self.distributions.add(node.id)
        if node.id in self.variables_lookup:
            self.variables.add(node.id)

prog = namedtuple('prog', ('imports', 'constants', 'variables', 'arrays', 'command'))
class SPPL_Compiler():
    def __init__(self, source, modelname='model'):
        self.source = source
        self.modelname = modelname
        self.prog = prog(
            imports=io.StringIO(),
            constants=io.StringIO(),
            variables=io.StringIO(),
            arrays=io.StringIO(),
            command=io.StringIO())
        self.compile()
    def preprocess(self):
        lines = self.source.split(os.linesep)
        source_prime = os.linesep.join(l for l in lines if l.strip())
        source_prime = source_prime.replace('~=', '=')
        source_prime = re.sub(
            r'^(\s*)switch\s*\((.+)\)\s*cases\s*\((.+)\s*in\s+(.+)\s*\)\s*:',
            r"\1for \3 in switch(\2, \4):",
            source_prime, flags=re.MULTILINE)
        return source_prime
    def compile(self):
        # Parse and visit.
        source_prime = self.preprocess()
        tree = ast.parse(source_prime)
        visitor = SPPL_Visitor()
        visitor.visit(tree)
        # Write the imports.
        self.prog.imports.write("# IMPORT STATEMENTS")
        self.prog.imports.write('\n')
        for i in visitor.imports:
            self.prog.imports.write(i)
            self.prog.imports.write('\n')
        for d in sorted(visitor.distributions):
            self.prog.imports.write('from sppl.distributions import %s' % (d,))
            self.prog.imports.write('\n')
        for c in ['Id', 'IdArray', 'Condition', 'IfElse', 'For', 'Sample',
                    'Sequence', 'Switch', 'Transform']:
            self.prog.imports.write('from sppl.compilers.ast_to_spn import %s' % (c,))
            self.prog.imports.write('\n')
        # Write the constants.
        if visitor.constants:
            self.prog.constants.write('# CONSTANT DECLRATIONS')
            self.prog.constants.write('\n')
            for c in visitor.constants:
                self.prog.constants.write(c)
                self.prog.constants.write('\n')
        # Write the variables.
        variables = [v for v, t in visitor.variables.items() if t[0] == 'variable']
        arrays = [(v, t[1]) for v, t in visitor.variables.items() if t[0] == 'array']
        if variables:
            self.prog.variables.write('# VARIABLE DECLARATIONS')
            self.prog.variables.write('\n')
            for v in variables:
                self.prog.variables.write('%s = Id(\'%s\')' % (v, v,))
                self.prog.variables.write('\n')
        if arrays:
            self.prog.arrays.write('# ARRAY DECLARATIONS')
            self.prog.arrays.write('\n')
            for v, n in arrays:
                self.prog.arrays.write('%s = IdArray(\'%s\', %s)' % (v, v, n,))
                self.prog.arrays.write('\n')
        # Write the command.
        self.prog.command.write('# MODEL DEFINITION')
        self.prog.command.write('\n')
        command = visitor.command.getvalue()
        self.prog.command.write('command = %s' % (command,))
        self.prog.command.write('\n')
        # Write the interpret step.
        self.prog.command.write('%s = command.interpret()' % (self.modelname,))
    def render_module(self):
        """Render the source code as a stand alone Python module."""
        program = io.StringIO()
        for stream in self.prog:
            v = stream.getvalue()
            if v:
                program.write(stream.getvalue())
                program.write('\n')
        return program.getvalue()
    def execute_module(self):
        """Execute the source code in a fresh module."""
        code = self.render_module()
        namespace = {}
        exec(code, namespace)
        return SimpleNamespace(**namespace)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import numpy
import pytest
import scipy.stats

from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.math_util import logdiffexp
from sppl.sets import Interval
from sppl.sets import Reals
from sppl.sets import inf as oo
from sppl.spn import ContinuousLeaf
from sppl.spn import SumSPN
from sppl.transforms import Id

def test_numeric_distribution_normal():
    X = Id('X')
    spn = (X >> norm(loc=0, scale=1))

    assert allclose(spn.logprob(X > 0), -log(2))
    assert allclose(spn.logprob(abs(X) < 2), log(spn.dist.cdf(2) - spn.dist.cdf(-2)))

    assert allclose(spn.logprob(X**2 > 0), 0)
    assert allclose(spn.logprob(abs(X) > 0), 0)
    assert allclose(spn.logprob(~(X << {1})), 0)

    assert isinf_neg(spn.logprob(X**2 - X + 10 < 0))
    assert isinf_neg(spn.logprob(abs(X) < 0))
    assert isinf_neg(spn.logprob(X << {1}))

    spn.sample(100)
    spn.sample_subset([X], 100)
    assert spn.sample_subset([], 100) == [{}]*100
    spn.sample_func(lambda X: X**2, 1)
    spn.sample_func(lambda X: abs(X)+X**2, 1)
    spn.sample_func(lambda X: X**2 if X > 0 else X**3, 100)

    spn_condition_a = spn.condition((X < 2) | (X > 10))
    samples = spn_condition_a.sample(100)
    assert all(s[X] < 2 for s in samples)

    spn_condition_b = spn.condition((X < -10) | (X > 10))
    assert isinstance(spn_condition_b, SumSPN)
    assert allclose(spn_condition_b.weights[0], -log(2))
    assert allclose(spn_condition_b.weights[0], spn_condition_b.weights[1])

    for event in [(X<-10), (X>3)]:
        spn_condition_c = spn.condition(event)
        assert isinstance(spn_condition_c, ContinuousLeaf)
        assert isinf_neg(spn_condition_c.logprob((-1 < X) < 1))
        samples = spn_condition_c.sample(100, prng=numpy.random.RandomState(1))
        assert all(s[X] in event.values for s in samples)

    with pytest.raises(ValueError):
        spn.condition((X > 1) & (X < 1))

    with pytest.raises(ValueError):
        spn.condition(X << {1})

    with pytest.raises(ValueError):
        spn.sample_func(lambda Z: Z**2, 1)

    x = spn.logprob((X << {1, 2}) | (X < -1))
    assert allclose(x, spn.logprob(X < -1))

    with pytest.raises(AssertionError):
        spn.logprob(Id('Y') << {1, 2})

def test_numeric_distribution_gamma():
    X = Id('X')

    spn = (X >> gamma(a=1, scale=1))
    with pytest.raises(ValueError):
        spn.condition((X << {1, 2}) | (X < 0))

    # Intentionally set Reals as the domain to exercise an important
    # code path in dist.condition (Union case with zero weights).
    spn = ContinuousLeaf(X, scipy.stats.gamma(a=1, scale=1), Reals)
    assert isinf_neg(spn.logprob((X << {1, 2}) | (X < 0)))
    with pytest.raises(ValueError):
        spn.condition((X << {1, 2}) | (X < 0))

    spn_condition = spn.condition((X << {1,2} | (X <= 3)))
    assert isinstance(spn_condition, ContinuousLeaf)
    assert spn_condition.conditioned
    assert spn_condition.support == Interval(-oo, 3)
    assert allclose(
        spn_condition.logprob(X <= 2),
        logdiffexp(spn.logprob(X<=2), spn.logprob(X<=0))
            - spn_condition.logZ)

    # Support on (-3, oo)
    spn = (X >> gamma(loc=-3, a=1))
    assert spn.prob((-3 < X) < 0) > 0.95



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import pytest

from sppl.distributions import choice
from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.distributions import poisson
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.math_util import logsumexp
from sppl.spn import SumSPN
from sppl.transforms import Id

X = Id('X')
Y = Id('Y')
Z = Id('Z')

def test_logpdf_real_continuous():
    spn = (X >> norm())
    assert allclose(spn.logpdf({X: 0}), norm().dist.logpdf(0))

def test_logpdf_real_discrete():
    spn = (X >> poisson(mu=2))
    assert isinf_neg(spn.logpdf({X: 1.5}))
    assert isinf_neg(spn.logpdf({X: '1'}))
    assert not isinf_neg(spn.logpdf({X: 0}))

def test_logpdf_nominal():
    spn = (X >> choice({'a' : .6, 'b': .4}))
    assert isinf_neg(spn.logpdf({X: 1.5}))
    allclose(spn.logpdf({X: 'a'}), log(.6))

def test_logpdf_mixture_real_continuous_continuous():
    spn = X >> (.3*norm() | .7*gamma(a=1))
    assert allclose(
        spn.logpdf({X: .5}),
        logsumexp([
            log(.3) + spn.children[0].logpdf({X: 0.5}),
            log(.7) + spn.children[1].logpdf({X: 0.5}),
        ]))

@pytest.mark.xfail
def test_logpdf_mixture_real_continuous_discrete():
    spn = X >> (.3*norm() | .7*poisson(mu=1))
    assert allclose(
        spn.logpdf(X << {.5}),
        logsumexp([
            log(.3) + spn.children[0].logpdf({X: 0.5}),
            log(.7) + spn.children[1].logpdf({X: 0.5}),
        ]))
    assert False, 'Invalid base measure addition'

def test_logpdf_mixture_nominal():
    spn = SumSPN([X >> norm(), X >> choice({'a':.1, 'b':.9})], [log(.4), log(.6)])
    assert allclose(
        spn.logpdf({X: .5}),
        log(.4) + spn.children[0].logpdf({X: .5}))
    assert allclose(
        spn.logpdf({X: 'a'}),
        log(.6) + spn.children[1].logpdf({X: 'a'}))

def test_logpdf_error_event():
    spn = (X >> norm())
    with pytest.raises(Exception):
        spn.logpdf(X < 1)

def test_logpdf_error_transform_base():
    spn = (X >> norm())
    with pytest.raises(Exception):
        spn.logpdf({X**2: 0})

def test_logpdf_error_transform_env():
    spn = (X >> norm()).transform(Z, X**2)
    with pytest.raises(Exception):
        spn.logpdf({Z: 0})

def test_logpdf_bivariate():
    spn = (X >> norm()) & (Y >> choice({'a': .5, 'b': .5}))
    assert allclose(
        spn.logpdf({X: 0, Y: 'a'}),
        norm().dist.logpdf(0) + log(.5))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

"""Convert AST to SPN."""

import os

from functools import reduce

from ..dnf import dnf_normalize
from ..math_util import allclose
from ..math_util import isinf_neg
from ..math_util import logsumexp
from ..sets import FiniteNominal
from ..sets import FiniteReal
from ..sets import Set
from ..spn import Memo
from ..spn import SumSPN
from ..spn import spn_simplify_sum

from .. import transforms

inf = float('inf')

Id = transforms.Id
def IdArray(token, n):
    return [Id('%s[%d]' % (token, i,)) for i in range(n)]

class Command():
    def interpret(self, spn):
        raise NotImplementedError()

class Skip(Command):
    def __init__(self):
        pass
    def interpret(self, spn=None):
        return spn

class Sample(Command):
    def __init__(self, symbol, distribution):
        self.symbol = symbol
        self.distribution = distribution
    def interpret(self, spn=None):
        leaf = self.symbol >> self.distribution
        return leaf if (spn is None) else spn & leaf

class Transform(Command):
    def __init__(self, symbol, expr):
        self.symbol = symbol
        self.expr = expr
    def interpret(self, spn=None):
        assert spn is not None
        return spn.transform(self.symbol, self.expr)

class Condition(Command):
    def __init__(self, event):
        self.event = event
    def interpret(self, spn=None):
        assert spn is not None
        return spn.condition(self.event)

class IfElse(Command):
    def __init__(self, *branches):
        assert len(branches) % 2 == 0
        self.branches = branches
    def interpret(self, spn=None):
        assert spn is not None
        conditions = self.branches[::2]
        subcommands = self.branches[1::2]
        # Make events for each condition.
        if conditions[-1] is True:
            events_if = [
                reduce(lambda x, e: x & ~e, conditions[:i], conditions[i])
                for i in range(len(conditions)-1)
            ]
            event_else = ~reduce(lambda x, e: x|e, conditions[:-1])
            events_unorm = events_if + [event_else]
        else:
            events_unorm = [
                reduce(lambda x, e: x & ~e, conditions[:i], conditions[i])
                for i in range(len(conditions))
            ]
        # Rewrite events in normalized form.
        events = [dnf_normalize(event) for event in events_unorm]
        # Rewrite events in normalized form.
        return interpret_if_block(spn, events, subcommands)

class For(Command):
    def __init__(self, n0, n1, f):
        self.n0 = n0
        self.n1 = n1
        self.f = f
    def interpret(self, spn=None):
        commands = [self.f(i) for i in range(self.n0, self.n1)]
        sequence = Sequence(*commands)
        return sequence.interpret(spn)

class Switch(Command):
    def __init__(self, symbol, values, f):
        self.symbol = symbol
        self.f = f
        self.values = values
    def interpret(self, spn=None):
        if isinstance(self.values, enumerate):
            values = list(self.values)
            sets = [self.value_to_set(v[1]) for v in values]
            subcommands = [self.f(*v) for v in values]
        else:
            sets = [self.value_to_set(v) for v in self.values]
            subcommands = [self.f(v) for v in self.values]
        sets_disjoint = [
            reduce(lambda x, s: x & ~s, sets[:i], sets[i])
            for i in range(len(sets))]
        events = [self.symbol << s for s in sets_disjoint]
        return interpret_if_block(spn, events, subcommands)
    def value_to_set(self, v):
        if isinstance(v, Set):
            return v
        if isinstance(v, str):
            return FiniteNominal(v)
        return FiniteReal(v)

class Sequence(Command):
    def __init__(self, *commands):
        self.commands = commands
    def interpret(self, spn=None):
        return reduce(lambda S, c: c.interpret(S), self.commands, spn)

Otherwise = True

def interpret_if_block(spn, events, subcommands):
    assert len(events) == len(subcommands)
    # Prepare memo table.
    memo = Memo()
    # Obtain mixture probabilities.
    weights = [spn.logprob(event, memo)
        if event is not None else -inf for event in events]
    # Filter the irrelevant ones.
    indexes = [i for i, w in enumerate(weights) if not isinf_neg(w)]
    assert indexes, 'All conditions probability zero.'
    # Obtain conditioned SPNs.
    weights_conditioned = [weights[i] for i in indexes]
    spns_conditioned = [spn.condition(events[i], memo) for i in indexes]
    subcommands_conditioned = [subcommands[i] for i in indexes]
    assert allclose(logsumexp(weights_conditioned), 0)
    # Make the children.
    children = [
        subcommand.interpret(S)
        for S, subcommand in zip(spns_conditioned, subcommands_conditioned)
    ]
    # Maybe Simplify.
    if len(children) == 1:
        spn = children[0]
    else:
        spn = SumSPN(children, weights_conditioned)
        if not os.environ.get('SPN_NO_SIMPLIFY'):
            spn = spn_simplify_sum(spn)
    # Return the SPN.
    return spn



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import numpy

from sppl.distributions import bernoulli
from sppl.distributions import choice
from sppl.distributions import norm
from sppl.spn import ProductSPN
from sppl.spn import SumSPN
from sppl.spn import spn_cache_duplicate_subtrees
from sppl.transforms import Id

rng = numpy.random.RandomState(1)

W = Id('W')
Y = Id('Y')
X = [Id('X[0]'), Id('X[1]')]
Z = [Id('Z[0]'), Id('Z[1]')]

def test_cache_simple_leaf():
    spn = .5 * (W >> norm(loc=0, scale=1)) | .5 * (W >> norm(loc=0, scale=1))
    assert spn.children[0] is not spn.children[1]
    spn_cached = spn_cache_duplicate_subtrees(spn, {})
    assert spn_cached.children[0] is spn_cached.children[1]

def test_cache_simple_sum_of_product():
    spn \
        = 0.3 * ((W >> norm(loc=0, scale=1)) & (Y >> norm(loc=0, scale=1))) \
        | 0.7 * ((W >> norm(loc=0, scale=1)) & (Y >> norm(loc=0, scale=2)))
    spn_cached = spn_cache_duplicate_subtrees(spn, {})
    assert spn_cached.children[0].children[0] is spn_cached.children[1].children[0]

def test_cache_complex_sum_of_product():
    # Test case adapted from the SPN generated by
    # test_repeat.make_model_repeat(n=2)
    duplicate_subtrees = [None, None]
    for i in range(2):
        duplicate_subtrees[i] = SumSPN([
            ProductSPN([
                (X[0] >> bernoulli(p=.1)),
                SumSPN([
                    (Z[0] >> bernoulli(p=.5))
                        & (Y >> choice({'0':.1, '1': .9})),
                    (Z[0] >> bernoulli(p=.1))
                        & (Y >> choice({'0':.9, '1': .1}))
                ], weights=[log(.730), log(.270)])
            ]),
            ProductSPN([
                Z[0] >> bernoulli(p=.1),
                Y >> choice({'0':.9, '1':.1}),
                X[0] >> bernoulli(p=.5),
            ]),
        ], weights=[log(.925), log(.075)])

    assert duplicate_subtrees[0] == duplicate_subtrees[1]
    assert duplicate_subtrees[0] is not duplicate_subtrees[1]

    left_subtree = ProductSPN([
        X[1] >> bernoulli(p=.5),
        SumSPN([
            ProductSPN([
                duplicate_subtrees[0],
                Z[1] >> bernoulli(p=.5),
            ]),
            ProductSPN([
                Z[1] >> bernoulli(p=.7),
                SumSPN([
                    Y >> choice({'0':.3, '1':.7})
                        & X[0] >> bernoulli(p=.1)
                        & Z[0] >> bernoulli(p=.1),
                    Y >> choice({'0':.7, '1':.3})
                        & X[0] >> bernoulli(p=.5)
                        & Z[0] >> bernoulli(p=.5),
                ], weights=[log(.9), log(.1)])
            ])
        ], weights=[log(.783), log(.217)])
    ])

    right_subtree = ProductSPN([
        Z[1] >> bernoulli(p=.8),
        X[1] >> bernoulli(p=.1),
        duplicate_subtrees[1]
    ])

    spn = .92 * left_subtree | .08 * right_subtree

    spn_cached = spn_cache_duplicate_subtrees(spn, {})
    assert spn_cached.children[0].children[1].children[0].children[0] is duplicate_subtrees[0]
    assert spn_cached.children[1].children[2] is duplicate_subtrees[0]



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

'''
Burglary network example from:

Artificial Intelligence: A Modern Approach (3rd Edition).
Russel and Norvig, Fig 14.2 pp 512.
'''

from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Otherwise
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.distributions import bernoulli

Burglary    = Id('Burglary')
Earthquake  = Id('Earthquake')
Alarm       = Id('Alarm')
JohnCalls   = Id('JohnCalls')
MaryCalls   = Id('MaryCalls')

program = Sequence(
    Sample(Burglary,   bernoulli(p=0.001)),
    Sample(Earthquake, bernoulli(p=0.002)),
    IfElse(
        Burglary << {1},
            IfElse(
                Earthquake << {1},  Sample(Alarm, bernoulli(p=0.95)),
                Otherwise,          Sample(Alarm, bernoulli(p=0.94))),
        Otherwise,
            IfElse(
                Earthquake << {1},  Sample(Alarm, bernoulli(p=0.29)),
                Otherwise,          Sample(Alarm, bernoulli(p=0.001)))),
    IfElse(
        Alarm << {1}, Sequence(
            Sample(JohnCalls, bernoulli(p=0.90)),
            Sample(MaryCalls, bernoulli(p=0.70))),
        Otherwise, Sequence(
                Sample(JohnCalls, bernoulli(p=0.05)),
                Sample(MaryCalls, bernoulli(p=0.01))),
        ))
model = program.interpret()

def test_marginal_probability():
    # Query on pp. 514.
    event = ((JohnCalls << {1}) & (MaryCalls << {1}) & (Alarm << {1})
        & (Burglary << {0}) & (Earthquake << {0}))
    x = model.prob(event)
    assert str(x)[:8] == '0.000628'

def test_conditional_probability():
    # Query on pp. 523
    event = (JohnCalls << {1}) & (MaryCalls << {1})
    model_condition = model.condition(event)
    x = model_condition.prob(Burglary << {1})
    assert str(x)[:5] == '0.284'

def test_mutual_information():
    event_a = (JohnCalls << {1}) | (MaryCalls << {1})
    event_b = (Burglary << {1}) & (Earthquake << {0})
    print(model.mutual_information(event_a, event_b))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import os
import re
from setuptools import setup

# Specify the requirements.
requirements = {
    'src' : [
        'astunparse==1.6.3',
        'numpy==1.16.6',
        'scipy==1.4.1',
        'sympy==1.6',
    ],
    'magics' : [
        'graphviz==0.13.2',
        'ipython==7.13.0',
        'jupyter-core==4.6.3',
        'networkx==2.4',
        'notebook==6.0.3',
        'matplotlib==3.3.2',
        'pygraphviz==1.5',
    ],
    'tests' : [
        'pytest-timeout==1.3.3',
        'pytest==5.2.2',
        'coverage==5.3',
    ]
}
requirements['all'] = [r for v in requirements.values() for r in v]

# Determine the version (hardcoded).
dirname = os.path.dirname(os.path.realpath(__file__))
vre = re.compile('__version__ = \'(.*?)\'')
m = open(os.path.join(dirname, 'src', '__init__.py')).read()
__version__ = vre.findall(m)[0]

setup(
    name='sppl',
    version=__version__,
    description='The Sum-Product Probabilistic Language',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/probcomp/sppl',
    license='Apache-2.0',
    maintainer='Feras A. Saad',
    maintainer_email='fsaad@mit.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
    ],
    packages=[
        'sppl',
        'sppl.compilers',
        'sppl.magics',
        'sppl.tests',
    ],
    package_dir={
        'sppl'           : 'src',
        'sppl.compilers' : 'src/compilers',
        'sppl.magics'    : 'magics',
        'sppl.tests'     : 'tests',
    },
    install_requires=requirements['src'],
    # Allow extras packaging.
    # install_requires=requirements['src'],
    extras_require=requirements,
    python_requires='>=3.6',
)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from functools import reduce
from itertools import chain

from .math_util import int_or_isinf_neg
from .math_util import int_or_isinf_pos
from .math_util import isinf_neg
from .math_util import isinf_pos

inf = float('inf')
oo = inf

class Set:
    pass

# EmptySetC shall have a single instance.
class EmptySetC(Set):
    def __init__(self, force=None):
        assert force
    def __contains__(self, x):
        return False
    def __invert__(self):
        # This case is tricky; by convention, we return Real line.
        # return Union(FiniteNominal(b=True), Interval(-inf, inf))
        return Interval(-inf, inf)
    def __and__(self, x):
        if isinstance(x, Set):
            return self
        return NotImplemented
    def __or__(self, x):
        if isinstance(x, Set):
            return x
        return NotImplemented
    def __eq__(self, x):
        return x is self
    def __hash__(self):
        x = (self.__class__,)
        return hash(x)
    def __repr__(self):
        return 'EmptySet'
    def __str__(self):
        return 'EmptySet'

class FiniteNominal(Set):
    def __init__(self, *values, b=None):
        assert values or b
        self.values = frozenset(values)
        self.b = b
    def __contains__(self, x):
        if self.b:
            return x not in self.values
        return x in self.values
    def __invert__(self):
        if not self.values:
            assert self.b
            return EmptySet
        return FiniteNominal(*self.values, b=not self.b)
    def __and__(self, x):
        if isinstance(x, FiniteNominal):
            if not self.b:
                values = {v for v in self.values if v in x}
                return FiniteNominal(*values) if values else EmptySet
            if not x.b:
                values = {v for v in x.values if v in self}
                return FiniteNominal(*values) if values else EmptySet
            values = self.values | x.values
            return FiniteNominal(*values, b=True)
        if isinstance(x, (FiniteReal, Interval)):
            return EmptySet
        if isinstance(x, Set):
            return x & self
        return NotImplemented
    def __or__(self, x):
        if isinstance(x, FiniteNominal):
            if self.b:
                values = {v for v in self.values if v not in x}
                return FiniteNominal(*values, b=self.b)
            if x.b:
                values = {v for v in x.values if v not in self}
                return FiniteNominal(*values, b=x.b)
            values = self.values | x.values
            return FiniteNominal(*values, b=False)
        if isinstance(x, (FiniteReal, Interval)):
            return Union(self, x)
        if isinstance(x, Set):
            return x | self
        return NotImplemented
    def __eq__(self, x):
        return isinstance(x, FiniteNominal) \
            and bool(self.b) == bool(x.b) \
            and self.values == x.values
    def __hash__(self):
        x = (self.__class__, self.values, self.b)
        return hash(x)
    def __repr__(self):
        str_values = ', '.join(repr(x) for x in self.values)
        return 'FiniteNominal(%s, b=%s)' % (str_values, repr(self.b))
    def __str__(self):
        return '%s%s' % (('~' if self.b else ''), str(set(self.values)),)
    def __len__(self):
        return len(self.values)
    def __iter__(self):
        return iter(sorted(self.values))

class FiniteReal(Set):
    def __init__(self, *values):
        assert values
        self.values = frozenset(values)
    def __contains__(self, x):
        # inf == oo but hash(inf) != hash(oo)
        return any(x == v for v in self.values)
    def __invert__(self):
        values = sorted(self.values)
        intervals = chain(
            # Left-infinity interval.
            [Interval.Ropen(-inf, values[0])],
            # Finite intervals.
            [Interval.open(x, y) for x, y in zip(values, values[1:])],
            # Right-infinity interval.
            [Interval.Lopen(values[-1], inf)])
        return Union(*intervals)
    def __and__(self, x):
        if isinstance(x, FiniteReal):
            values = self.values & x.values
            return FiniteReal(*values) if values else EmptySet
        if isinstance(x, Interval):
            values = {v for v in self.values if v in x}
            return FiniteReal(*values) if values else EmptySet
        if isinstance(x, FiniteNominal):
            return EmptySet
        if isinstance(x, Set):
            return x & self
        return NotImplemented
    def __or__(self, x):
        if isinstance(x, FiniteReal):
            values = self.values | x.values
            return FiniteReal(*values)
        if isinstance(x, Interval):
            # Merge endpoints.
            values = set(self.values)
            interval = x
            if interval.a in values and interval.left_open:
                values.remove(interval.a)
                interval = Interval(interval.a, interval.b,
                    left_open=None,
                    right_open=interval.right_open)
            if interval.b in values and interval.right_open:
                values.remove(interval.b)
                interval = Interval(interval.a, interval.b,
                    left_open=interval.left_open,
                    right_open=None)
            values = {v for v in values if v not in interval}
            return Union(FiniteReal(*values), interval) if values else interval
        if isinstance(x, FiniteNominal):
            return Union(self, x)
        if isinstance(x, Set):
            return x | self
        return NotImplemented
    def __eq__(self, x):
        return isinstance(x, FiniteReal) \
            and self.values == x.values
    def __hash__(self):
        x = (self.__class__, self.values)
        return hash(x)
    def __repr__(self):
        return 'FiniteReal(%s)' % (', '.join(repr(x) for x in self.values))
    def __str__(self):
        return str(set(self.values))
    def __len__(self):
        return len(self.values)
    def __iter__(self):
        return iter(sorted(self.values))

class Interval(Set):
    def __init__(self, a, b, left_open=None, right_open=None):
        assert a < b
        self.a = a
        self.b = b
        self.left_open = left_open or isinf_neg(self.a)
        self.right_open = right_open or isinf_pos(self.b)
        # SymPy compatibility.
        (self.left, self.right) = (a, b)
    def __contains__(self, x):
        if self.left_open and self.right_open:
            return self.a < x < self.b
        if self.left_open and not self.right_open:
            return self.a < x <= self.b
        if not self.left_open and self.right_open:
            return self.a <= x < self.b
        if not self.left_open and not self.right_open:
            return self.a <= x <= self.b
        assert False
    def __invert__(self):
        if isinf_neg(self.a):
            if isinf_pos(self.b):
                return EmptySet
            return Interval(self.b, inf, left_open=not self.right_open)
        if isinf_pos(self.b):
            return Interval(-inf, self.a, right_open=not self.left_open)
        left = Interval(-inf, self.a, right_open=not self.left_open)
        right = Interval(self.b, inf, left_open=not self.right_open)
        return Union(left, right)
    def __and__(self, x):
        if isinstance(x, Interval):
            if x == self:
                return x
            if (x.a in self) and (x.b in self):
                return x
            if (self.a in x) and (self.b in x):
                return self
            if x.a in self:
                if self.b == x.a:
                    return FiniteReal(x.a) if x.a in x else EmptySet
                return Interval(x.a, self.b, left_open=x.left_open, right_open=self.right_open)
            if x.b in self:
                if self.a == x.b:
                    return FiniteReal(x.b) if x.b in x else EmptySet
                return Interval(self.a, x.b, left_open=self.left_open, right_open=x.right_open)
            if self.a == x.a:
                left_open = self.left_open or x.left_open
                return Interval(self.a, self.b, left_open=left_open, right_open=self.right_open)
            if self.b == x.b:
                right_open = self.right_open or x.right_open
                return Interval(self.a, self.b, left_open=self.left_open, right_open=right_open)
            return EmptySet
        if isinstance(x, Set):
            return x & self
        return NotImplemented
    def __or__(self, x):
        if isinstance(x, Interval):
            if self == x:
                return self
            intersection = x & self
            if intersection is EmptySet \
                    and (self.a not in x) \
                    and (self.b not in x) \
                    and (x.a not in self) \
                    and (x.b not in self):
                return Union(self, x)
            (al, am, bm, br) = sorted((
                (self.a, self.left_open),
                (self.b, self.right_open),
                (x.a, x.left_open),
                (x.b, x.right_open)))
            left_open = al[1] if al[0] < am[0] else (al[1] and am[1])
            right_open = br[1] if bm[0] < br[0] else (bm[1] and br[1])
            return Interval(al[0], br[0], left_open=left_open, right_open=right_open)
        if isinstance(x, Set):
            return x | self
        return NotImplemented
    def __eq__(self, x):
        return isinstance(x, Interval) \
            and self.a == x.a \
            and self.b == x.b \
            and bool(self.left_open) == bool(x.left_open) \
            and bool(self.right_open) == bool(x.right_open)
    def __hash__(self):
        x = (self.__class__, self.a, self.b, self.left_open, self.right_open)
        return hash(x)
    def __repr__(self):
        return 'Interval(%s, %s, left_open=%s, right_open=%s)' \
            % (repr(self.a), repr(self.b), repr(self.left_open), repr(self.right_open))
    def __str__(self):
        lp = '(' if self.left_open else '['
        rp = ')' if self.right_open else ']'
        return '%s%s,%s%s' % (lp, self.a, self.b, rp)
    @staticmethod
    def Lopen(a, b):
        return Interval(a, b, left_open=True)
    @staticmethod
    def Ropen(a, b):
        return Interval(a, b, right_open=True)
    @staticmethod
    def open(a, b):
        return Interval(a, b, left_open=True, right_open=True)

class Union(Set):
    def __init__(self, *values):
        # Do not use the constructor directly;
        # instead use the Python "or" operator.
        assert all(not isinstance(x, Union) for x in values)
        valuesne = [x for x in values if x is not EmptySet]
        assert valuesne
        nominals = [x for x in valuesne if isinstance(x, FiniteNominal)]
        atoms = [x for x in valuesne if isinstance(x, FiniteReal)]
        assert len(nominals) <= 1
        assert len(atoms) <= 1
        self.nominals = nominals[0] if nominals else EmptySet
        self.atoms = atoms[0] if atoms else EmptySet
        self.intervals = frozenset(x for x in valuesne if isinstance(x, Interval))
        # Build the values.
        vals = []
        if nominals:
            vals.append(self.nominals)
        if atoms:
            vals.append(self.atoms)
        for i in self.intervals:
            vals.append(i)
        self.values = frozenset(vals)
        assert 2 <= len(self.values)
        # SymPy compatibility
        self.args = valuesne
    def __contains__(self, x):
        return any(x in v for v in self.values)
    def __eq__(self, x):
        return isinstance(x, Union) \
            and self.values == x.values
    def __hash__(self):
        x = (self.__class__, self.values)
        return hash(x)
    def __repr__(self):
        return 'Union(%s)' % (', '.join(repr(v) for v in self.args))
    def __str__(self):
        return 'Union(%s)' % (', '.join(str(v) for v in self.args))
    def __and__(self, x):
        if x is EmptySet:
            return EmptySet
        if isinstance(x, FiniteNominal):
            return self.nominals & x
        if isinstance(x, (FiniteReal, Interval)):
            atoms = self.atoms & x
            intervals = [i & x for i in self.intervals]
            intervalsne = [i for i in intervals if i is not EmptySet]
            if atoms is EmptySet:
                if not intervalsne:
                    return EmptySet
                if len(intervalsne) == 1:
                    return intervalsne[0]
            if not intervalsne:
                return atoms
            return Union(atoms, *intervalsne)
        if isinstance(x, Union):
            terms = [self & v for v in x.values]
            return reduce(lambda a,b: a |b, terms)
    def __or__(self, x):
        if x is EmptySet:
            return self
        if isinstance(x, FiniteNominal):
            nominals = self.nominals | x
            return Union(nominals, self.atoms, *self.intervals)
        if isinstance(x, FiniteReal):
            atoms = self.atoms | x
            blocks = union_intervals_finite(self.intervals, atoms)
            assert blocks
            if len(blocks) == 1 and self.nominals is EmptySet:
                return blocks[0]
            return Union(self.nominals, *blocks)
        if isinstance(x, Interval):
            intervals = list(self.intervals) + [x]
            blocks = union_intervals_finite(intervals, self.atoms)
            assert blocks
            if len(blocks) == 1 and self.nominals is EmptySet:
                return blocks[0]
            return Union(self.nominals, *blocks)
        if isinstance(x, Union):
            return reduce(lambda a,b: a | b, x.values, self)
        return NotImplemented
    def __invert__(self):
        inversions = [~x for x in self.values]
        return reduce(lambda a,b: a&b, inversions)
    def __iter__(self):
        return iter(self.args)

def union_intervals(intervals):
    intervals_sorted = sorted(intervals, key=lambda i:i.a)
    blocks = [intervals_sorted[0]]
    for interval in intervals_sorted[1:]:
        interval_union = blocks[-1] | interval
        if isinstance(interval_union, Interval):
            blocks[-1] = interval_union
        elif isinstance(interval_union, Union):
            blocks.append(interval)
        else:
            assert False
    return blocks

def union_intervals_finite(intervals, finite):
    if finite is EmptySet:
        return union_intervals(intervals)
    blocks = []
    finite_current = finite
    for interval in intervals:
        interval_union = interval | finite_current
        if isinstance(interval_union, Interval):
            blocks.append(interval_union)
            finite_current = EmptySet
        elif isinstance(interval_union, Union):
            assert interval_union.atoms is not EmptySet
            assert len(interval_union.intervals) == 1
            interval_part = next(iter(interval_union.intervals))
            blocks.append(interval_part)
            finite_current = interval_union.atoms
        else:
            assert False
    blocks_merged = union_intervals(blocks)
    if finite_current is not EmptySet:
        blocks_merged.append(finite_current)
    return blocks_merged

def make_union(*args):
    return reduce(lambda a,b: a|b, args)
def make_intersection(*args):
    return reduce(lambda a,b: a&b, args)

EmptySet = EmptySetC(force=1)
Reals = Interval(-inf, inf)
RealsPos = Interval(0, inf)
RealsNeg = Interval(-inf, 0)
ExtReals = Union(FiniteReal(-inf, inf), Reals)
ExtRealsPos = Union(FiniteReal(inf), RealsPos)
Strings = FiniteNominal(b=True)

# Integral hacks.
Integers = Reals
IntegersPos = Interval.Lopen(0, inf)
IntegersPos0 = RealsPos
def Range(start, stop):
    assert int_or_isinf_neg(start)
    assert int_or_isinf_pos(stop)
    right_open = start == stop
    return Interval(start, stop + right_open, right_open=(start == stop))

# TODO: Expunge.
def convert_sympy(x):
    import sympy
    if x is sympy.S.EmptySet:
        return EmptySet
    if isinstance(x, sympy.Interval):
        left = -inf if x.left == -inf else x.left
        right = inf if x.right == -inf else x.right
        return Interval(left, right, left_open=x.left_open, right_open=x.right_open)
    if isinstance(x, sympy.FiniteSet):
        return FiniteReal(*x.args)
    if isinstance(x, sympy.Union):
        xs = [convert_sympy(v) for v in x.args]
        return make_union(*xs)
    assert False, 'Unknown set in sympy conversion: %s' % (x,)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.ast_to_spn import Transform
from sppl.distributions import bernoulli
from sppl.distributions import randint
from sppl.math_util import allclose

Y = Id('Y')
X = Id('X')

def test_ifelse_zero_conditions():
    command = Sequence(
        Sample(Y, randint(low=0, high=3)),
        IfElse(
            Y << {-1},  Transform(X, Y**(-1)),
            Y << {0},   Sample(X, bernoulli(p=1)),
            Y << {1},   Transform(X, Y),
            Y << {2},   Transform(X, Y**2),
            Y << {3},   Transform(X, Y**3),
        ))
    model = command.interpret()
    assert len(model.children) == 3
    assert len(model.weights) == 3
    assert allclose(model.weights[0], model.logprob(Y << {0}))
    assert allclose(model.weights[1], model.logprob(Y << {1}))
    assert allclose(model.weights[2], model.logprob(Y << {2}))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import json
import pytest

from sympy import sqrt

from sppl.compilers.spn_to_dict import spn_from_dict
from sppl.compilers.spn_to_dict import spn_to_dict
from sppl.distributions import choice
from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.distributions import poisson
from sppl.sets import EmptySet
from sppl.transforms import EventFiniteNominal
from sppl.transforms import Exp
from sppl.transforms import Exponential
from sppl.transforms import Id
from sppl.transforms import Log
from sppl.transforms import Logarithm

X = Id('X')
Y = Id('Y')

spns = [
    X >> norm(loc=0, scale=1),
    X >> poisson(mu=7),
    Y >> choice({'a': 0.5, 'b': 0.5}),
    (X >> norm(loc=0, scale=1)) & (Y >> gamma(a=1)),
    0.2*(X >> norm(loc=0, scale=1)) | 0.8*(X >> gamma(a=1)),
]
@pytest.mark.parametrize('spn', spns)
def test_serialize_equal(spn):
    metadata = spn_to_dict(spn)
    spn_json_encoded = json.dumps(metadata)
    spn_json_decoded = json.loads(spn_json_encoded)
    spn2 = spn_from_dict(spn_json_decoded)
    assert spn2 == spn

transforms = [
    X,
    X**(1,3),
    Exponential(X, base=3),
    Logarithm(X, base=2),
    2**Log(X),
    1/Exp(X),
    abs(X),
    1/X,
    2*X + X**3,
    (X/2)*(X<0) + (X**(1,2))*(0<=X),
    X < sqrt(3),
    X << [],
    ~(X << []),
    EventFiniteNominal(1/X**(1,10), EmptySet),
    X << {1, 2},
    X << {'a', 'x'},
    ~(X << {'a', '1'}),
    (X < 3) | (X << {1,2}),
    (X < 3) & (X << {1,2}),
]
@pytest.mark.parametrize('transform', transforms)
def test_serialize_env(transform):
    spn = (X >> norm()).transform(Y, transform)
    metadata = spn_to_dict(spn)
    spn_json_encoded = json.dumps(metadata)
    spn_json_decoded = json.loads(spn_json_encoded)
    spn2 = spn_from_dict(spn_json_decoded)
    assert spn2 == spn
